#!/usr/bin/perl -w
# IPFREEZE v 0.4.8
# Copyright (C) 2003 gregoire HUBERT.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Author : gregoire HUBERT <greg@coolkeums.org>

use strict;
use IPTables::IPv4::IPQueue qw(:constants);
use NetPacket::IP;
use NetPacket::UDP;
use NetPacket::TCP;
use AppConfig;

use constant TIMEOUT => 1000000;
use constant VERSION => "0.4.8";

use sigtrap qw(handler release_handler INT QUIT);
use sigtrap qw(handler hup HUP);

my $packet;
my $ip;
my $tcp;
my $udp;
my $proto;
my $src_ip_addr;
my $dst_port;
my $return;
my %blacklist;
my @whitelist;
my $flag=0;
my $localtime;
my $queue;
my $cmd_line_options = AppConfig->new({CASE=>1});
my $logfile;
my $logblacklist;
my $target;


# Dump help out and exit
sub give_help {
	print "Ipfreeze ".VERSION."\n";
	print "Copyrights Gregoire Hubert 02-2004\n\n";
	print "Usage : ipfreeze.pl [-h|--help] [-l|--logfile file] [-L|--loglevel level] [-b|--blacklist file] [-w|--whitelist file] [-t|--target] [-d|--duration] [-V|--version]\n";
	print "-l|--logfile : Use the given log file (default is /var/log/ipfreeze.log)\n";
	print "-L|--loglevel : Set the netfilter logs on : 0=none, 1=temporarily blacklisted, 2=all blacklisted (default is 1)\n";
	print "-b|--blacklist : Use the given blacklist file (default is /var/log/ipfreeze/ip_blacklist)\n";
	print "-w|--whitelist : Use the given whitelist file (default is /var/log/ipfreeze/ip_whitelist)\n";
	print "-t|--target : Specify the netfilter target for blacklisted hosts (default is DROP). It can be a user created chain.\n";
	print "-d|--duration : Blacklist period in seconds (default is 600)\n";
	print "-V|--version : Show the version and exit\n";
	print "-h|--help : show this message and exit\n";
	exit 1;
}

# Print the version out

 sub give_version {
 	print VERSION."\n";
	exit 1;
}
# Clean exit procedure to release the queue handler correctly

sub release_handler {
	undef $queue;
	open MONFIC, ">>$logfile";
	$localtime = time();
	print MONFIC localtime($localtime)." @@ SIGINT or SIGQUIT received ... exiting @@\n";
	close MONFIC;
	exit(0);
	}

# Seem that we received a packet

sub update {
	$return=`iptables -F blacklist;`;
	foreach (keys %blacklist) {
		if ($logblacklist) {
			if ($logblacklist==1) {	
				if ($blacklist{$_}>0) {
					$return=`iptables -A blacklist -s $_ -j LOG --log-prefix "Blacklisted :";`;
					}
				}
			elsif ($logblacklist==2) {
				$return=`iptables -A blacklist -s $_ -j LOG --log-prefix "Blacklisted :";`;
				}
			}
		$return=`iptables -A blacklist -s $_ -j $target;`;
		}
	}
	

# Reset and load configuration from files

sub hup {
	open MONFIC, ">>$logfile";
	$localtime = time();
	print MONFIC localtime($localtime)." @@ [re]loading configuration... @@\n";
	undef %blacklist;
	undef @whitelist;
	open WHTLST, "<".$cmd_line_options->whitelist();
	open BLKLST, "<".$cmd_line_options->blacklist();
	$localtime = time();
	while ($return=<WHTLST>) { 
		chomp $return;
		next if ((length $return) < 6);  
		$whitelist[++$#whitelist]=$return;
		}	
	close WHTLST;
	while ($return=<BLKLST>) {
		chomp $return;
		next if ((length $return) < 6);
		$blacklist{$return}=-20;
		}
	close BLKLST;
	close MONFIC;
	&update();
	}
	
$cmd_line_options->define("logfile|l=s");
$cmd_line_options->logfile("/var/log/ipfreeze.log");
$cmd_line_options->define("loglevel|L=s");
$cmd_line_options->loglevel(1);
$cmd_line_options->define("blacklist|b=s");
$cmd_line_options->blacklist("/etc/ipfreeze/ip_blacklist");
$cmd_line_options->define("whitelist|w=s");
$cmd_line_options->whitelist("/etc/ipfreeze/ip_whitelist");
$cmd_line_options->define("target|t=s");
$cmd_line_options->target("DROP");
$cmd_line_options->define("duration|d=s");
$cmd_line_options->duration(600);
$cmd_line_options->define("version|V!");
$cmd_line_options->define("help|h!");
$cmd_line_options->args();

give_help() if ($cmd_line_options->help());
give_version if ($cmd_line_options->version());

$logfile = $cmd_line_options->logfile();
$logblacklist = $cmd_line_options->loglevel();
$target = $cmd_line_options->target();

printf("starting firewall blacklist manager v ".VERSION." copyright 2004 - Grgoire HUBERT\n");
$queue = new IPTables::IPv4::IPQueue(copy_mode  => IPQ_COPY_PACKET,
					copy_range => 65535) 
	    or die IPTables::IPv4::IPQueue->errstr;
open MONFIC, ">>$logfile";
$localtime = time();
print MONFIC localtime($localtime)." \@\@ Start daemon ipfreeze ".VERSION.".\@\@\n";
close MONFIC;
&hup();
while (1) {
  if (!defined($packet = $queue->get_message(TIMEOUT))) {
  	    if (IPTables::IPv4::IPQueue->errstr=~/Timeout/) {
	    	foreach (keys %blacklist) {
			if (($blacklist{$_}>0) && (!(--$blacklist{$_}))) {
				delete  $blacklist{$_};
				open MONFIC, ">>$logfile";
				$localtime = time();
				print MONFIC localtime($localtime)." -- $_\n";
				close MONFIC;
				$flag=1;
				}
	    		}
	    	if ($flag) {
	    		&update();
			$flag=0;
			}
		}
	    else {
	    	die("ERREUR : '".IPTables::IPv4::IPQueue->errstr."'");
		}
	    }
  else {	    
	  $ip = NetPacket::IP->decode($packet->payload());
	  $src_ip_addr = $ip->{src_ip};
	  if ($ip->{proto}==6) {
		  $tcp = NetPacket::TCP->decode($ip->{data});
		  $dst_port = $tcp->{dest_port};
		  $proto="TCP";
		  }
	  else {
	  	if ($ip->{proto}==17) {
		  $udp = NetPacket::UDP->decode($ip->{data});
		  $dst_port = $udp->{dest_port};
		  $proto="UDP";
		  }
		else {
			if ($ip->{proto}==1) {
				$proto="ICMP";
				$dst_port="none";
				}
			else {
				$proto="unknown";
				$dst_port="unknown";
				}
			}
		}
	  if (!($return = grep {/$src_ip_addr/} @whitelist)) {
	  	$blacklist{$src_ip_addr}=$cmd_line_options->duration();
 		&update(); 
		open MONFIC, ">>$logfile";
		$localtime = time();
		print MONFIC localtime($localtime)." ++ $src_ip_addr (dst port=$dst_port protocol=$proto).\n";
		close MONFIC;
		}
	  else {
		open MONFIC, ">>$logfile";
		$localtime = time();
		print MONFIC localtime($localtime)." == Got packet from whitelisted address $src_ip_addr (dst_port=$dst_port protocol=$proto).\n";
		close MONFIC;
		}
	  $queue->set_verdict($packet->packet_id(), NF_DROP)
	  }
  }

